# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import operator
import typing

import pylibastcanopy as bindings

CXX_OP_TO_PYTHON_OP = {
    "+": [operator.pos, operator.add],
    "-": [operator.neg, operator.sub],
    "*": operator.mul,
    "/": operator.truediv,
    "%": operator.mod,
    "&": operator.and_,
    "|": operator.or_,
    "^": operator.xor,
    "!": operator.not_,
    "~": operator.inv,
    "<<": operator.lshift,
    ">>": operator.rshift,
    "==": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "+=": operator.iadd,
    "-=": operator.isub,
    "*=": operator.imul,
    "/=": operator.itruediv,
    "%=": operator.imod,
    "&=": operator.iand,
    "|=": operator.ior,
    "^=": operator.ixor,
    "<<=": operator.ilshift,
    ">>=": operator.irshift,
    "=": None,  # assignment in python can be indexed assignment or sliced assignment, but copy assignment doesn't exist.
}


class Function:
    """
    Represents a C++ function.

    For C++ operators types:
    https://en.cppreference.com/w/cpp/language/operators.
    """

    def __init__(
        self,
        name: str,
        return_type: bindings.Type,
        params: list[bindings.ParamVar],
        exec_space: bindings.execution_space,
    ):
        self.name = name
        self.return_type = return_type
        self.params = params
        self.is_operator = self.name.startswith("operator")
        self._op_str = self.name[8:] if self.is_operator else None
        self.exec_space = exec_space

    def __str__(self):
        return f"{self.name}({', '.join(str(p) for p in self.params)}) -> {self.return_type}"

    def __repr__(self):
        old = super().__repr__()
        return f"{old[:-1]} {self.__str__()}>"

    @property
    def mangled_name(self):
        name = self.name
        if self.is_overloaded_operator():
            py_op = self.overloaded_operator_to_python_operator
            name = "operator" + "_" + py_op.__name__
        name = name.replace(" ", "_")
        return name

    @property
    def param_types(self) -> list[bindings.Type]:
        return [a.type_ for a in self.params]

    def is_allocation_operator(self) -> bool:
        if self._op_str is None:
            return False
        return self._op_str.startswith("new")

    def is_deallocation_operator(self) -> bool:
        if self._op_str is None:
            return False
        return self._op_str.startswith("delete")

    def is_user_defined_literal(self) -> bool:
        if self._op_str is None:
            return False
        return self._op_str.startswith('""')

    def is_cowait_operator(self) -> bool:
        if self._op_str is None:
            return False
        return self._op_str.startswith("co_await")

    def is_overloaded_operator(self) -> bool:
        if self._op_str is None:
            return False
        return self._op_str in CXX_OP_TO_PYTHON_OP

    def is_copy_assignment_operator(self) -> bool:
        return self._op_str == "="

    def is_conversion_operator(self) -> bool:
        # This standalone function doesn't keep track of all types in C++
        # system, so we find out if it's a conversion operator by eliminating
        # other possibilities.
        return (
            self.is_operator
            and not self.is_overloaded_operator()
            and not self.is_allocation_operator()
            and not self.is_deallocation_operator()
            and not self.is_user_defined_literal()
            and not self.is_cowait_operator()
        )

    @property
    def overloaded_operator_to_python_operator(self):
        if not self.is_overloaded_operator():
            raise ValueError(f"{self.name} is not an overloaded operator")
        self._op_str = typing.cast(str, self._op_str)

        if self._op_str in ("+", "-"):
            # Note this is standalone function
            # T T::operator-(const T2 &lh, const T2 &rh) const;
            # T T::operator-(const T2 &x) const;
            return CXX_OP_TO_PYTHON_OP[self._op_str][len(self.params) - 1]
        else:
            return CXX_OP_TO_PYTHON_OP[self._op_str]

    @classmethod
    def from_c_obj(cls, c_obj: bindings.Function):
        return cls(c_obj.name, c_obj.return_type, c_obj.params, c_obj.exec_space)


class FunctionTemplate:
    def __init__(
        self,
        template_parameters: list[bindings.TemplateParam],
        function: Function,
        num_min_required_args: int,
    ):
        self.template_parameters = template_parameters
        self.function = function
        self.num_min_required_args = num_min_required_args

    @classmethod
    def from_c_obj(cls, c_obj: bindings.FunctionTemplate):
        return cls(
            c_obj.template_parameters,
            Function.from_c_obj(c_obj.function),
            c_obj.num_min_required_args,
        )


class StructMethod(Function):
    def __init__(
        self,
        name: str,
        return_type: bindings.Type,
        params: list[bindings.ParamVar],
        kind: bindings.method_kind,
        exec_space: bindings.execution_space,
        is_move_constructor: bool = False,
    ):
        super().__init__(name, return_type, params, exec_space)
        self.kind = kind
        self.is_move_constructor = is_move_constructor

    @property
    def overloaded_operator_to_python_operator(self):
        if not self.is_overloaded_operator():
            raise ValueError(f"{self.name} is not an overloaded operator")
        self._op_str = typing.cast(str, self._op_str)

        if self._op_str in ("-", "+"):
            # Note that this is inside class definition, not standalone
            # T T::operator-() const;
            # T T::operator-(const T2 &b) const;
            return CXX_OP_TO_PYTHON_OP[self._op_str][len(self.params)]
        else:
            return CXX_OP_TO_PYTHON_OP[self._op_str]

    @classmethod
    def from_c_obj(cls, c_obj: bindings.Method):
        return cls(
            c_obj.name,
            c_obj.return_type,
            c_obj.params,
            c_obj.kind,
            c_obj.exec_space,
            c_obj.is_move_constructor(),
        )


class TemplatedStructMethod(StructMethod):
    @property
    def decl_name(self):
        """For a templated struct method, if the name contains template parameters,
        Decl name is the name without the template parameters. e.g.

        template<typename T, int n>
        struct Foo { Foo() {} };

        The name of the constructor is Foo<T, n>. Decl name is Foo.
        """

        if "<" in self.name:
            return self.name.split("<")[0]
        else:
            return self.name


class Struct:
    def __init__(
        self,
        name: str,
        fields: list[bindings.Field],
        methods: list[StructMethod],
        templated_methods: list[FunctionTemplate],
        nested_records: list[bindings.Record],
        nested_class_templates: list[bindings.ClassTemplate],
        sizeof_: int,
        alignof_: int,
    ):
        self.name = name
        self.fields = fields
        self.methods = methods
        self.templated_methods = templated_methods
        self.nested_records = nested_records
        self.nested_class_templates = nested_class_templates
        self.sizeof_ = sizeof_
        self.alignof_ = alignof_

    def constructors(self):
        for m in self.methods:
            if m.name == self.name:
                yield m

    def overloaded_operators(self):
        for m in self.methods:
            if m.is_overloaded_operator():
                yield m

    def conversion_operators(self):
        for m in self.methods:
            if m.is_conversion_operator():
                yield m

    @classmethod
    def from_c_obj(cls, c_obj: bindings.Record):
        methods = [StructMethod.from_c_obj(m) for m in c_obj.methods]
        return cls(
            c_obj.name,
            c_obj.fields,
            methods,
            [FunctionTemplate.from_c_obj(tm) for tm in c_obj.templated_methods],
            c_obj.nested_records,
            c_obj.nested_class_templates,
            c_obj.sizeof_,
            c_obj.alignof_,
        )


class TemplatedStruct(Struct):
    def __init__(
        self,
        name: str,
        fields: list[bindings.Field],
        methods: list[TemplatedStructMethod],
        templated_methods: list[FunctionTemplate],
        nested_records: list[bindings.Record],
        nested_class_templates: list[bindings.ClassTemplate],
        sizeof_: int,
        alignof_: int,
    ):
        super().__init__(
            name,
            fields,
            [],
            templated_methods,
            nested_records,
            nested_class_templates,
            sizeof_,
            alignof_,
        )
        self.methods = methods

    @classmethod
    def from_c_obj(cls, c_obj: bindings.Record):
        methods = [TemplatedStructMethod.from_c_obj(m) for m in c_obj.methods]

        return cls(
            c_obj.name,
            c_obj.fields,
            methods,
            [FunctionTemplate.from_c_obj(tm) for tm in c_obj.templated_methods],
            c_obj.nested_records,
            c_obj.nested_class_templates,
            c_obj.sizeof_,
            c_obj.alignof_,
        )


class ClassTemplate:
    def __init__(
        self,
        record: bindings.Record,
        template_parameters: list[bindings.TemplateParam],
        num_min_required_args: int,
    ):
        self.record = TemplatedStruct.from_c_obj(record)
        self.template_parameters = template_parameters
        self.num_min_required_args = num_min_required_args

    @classmethod
    def from_c_obj(cls, c_obj: bindings.ClassTemplate):
        return cls(c_obj.record, c_obj.template_parameters, c_obj.num_min_required_args)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from enum import IntEnum
from tempfile import TemporaryFile
import os
from contextlib import contextmanager


class STREAMFD(IntEnum):
    STDIN = 0
    STDOUT = 1
    STDERR = 2


class MinFDCapture:
    """Captures a file descriptor by redirecting it to a temporary file.
    Note that this is a minimum implementation. Corner cases may exist.
    for example, it currently does not handle stdin redirection.
    """

    def __init__(self, fd: STREAMFD):
        if fd == STREAMFD.STDIN:
            raise NotImplementedError("Cannot capture stdin")

        self.fd = fd
        self.old_fd = os.dup(fd)
        self.tmpfile = TemporaryFile()
        self.tmpfile_fd = self.tmpfile.fileno()

    def start(self):
        """Start capturing by redirecting the file descriptor to a temporary file."""
        try:
            os.fstat(self.old_fd)
        except (AttributeError, OSError) as e:
            raise ValueError("Old file descriptor is invalid") from e
        finally:
            os.dup2(self.tmpfile_fd, self.fd)

    def stop(self):
        """Restore the file descriptor with the saved descriptor."""
        os.dup2(self.old_fd, self.fd)

    def snap(self) -> str:
        """Get the captured content as a string."""
        self.tmpfile.seek(0)
        return self.tmpfile.read().decode("UTF-8")


@contextmanager
def capture_fd(fd: STREAMFD):
    """Context manager that captures a file descriptor and restores it after the block."""
    capturer = MinFDCapture(fd)
    capturer.start()
    try:
        yield capturer
    finally:
        capturer.stop()



from _typeshed import Incomplete
from typing import ClassVar, overload

class ClassTemplate(Template):
    num_min_required_args: int
    record: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...

class Declarations:
    class_templates: list[ClassTemplate]
    enums: list[Enum]
    function_templates: list[FunctionTemplate]
    functions: list[Function]
    records: list[Record]
    typedefs: list[Typedef]
    def __init__(self, *args, **kwargs) -> None: ...

class Enum:
    enumerator_values: list[str]
    enumerators: list[str]
    name: str
    def __init__(self, arg0) -> None: ...

class Field:
    access: access_kind
    name: str
    type_: Type
    def __init__(self, *args, **kwargs) -> None: ...

class Function:
    exec_space: execution_space
    name: str
    params: list[ParamVar]
    return_type: Type
    def __init__(self, *args, **kwargs) -> None: ...

class FunctionTemplate(Template):
    function: Function
    num_min_required_args: int
    def __init__(self, *args, **kwargs) -> None: ...

class Method(Function):
    kind: method_kind
    def __init__(self, *args, **kwargs) -> None: ...
    def is_move_constructor(self) -> bool: ...

class ParamVar:
    name: str
    type_: Type
    def __init__(self, arg0: str, arg1: Type) -> None: ...

class Record:
    alignof_: int
    fields: list[Field]
    methods: list[Method]
    name: str
    nested_class_templates: list[ClassTemplate]
    nested_records: list[Record]
    sizeof_: int
    templated_methods: list[FunctionTemplate]
    def __init__(self, *args, **kwargs) -> None: ...

class Template:
    template_parameters: list[TemplateParam]
    def __init__(self, *args, **kwargs) -> None: ...

class TemplateParam:
    kind: template_param_kind
    name: str
    type_: Type
    def __init__(self, *args, **kwargs) -> None: ...

class Type:
    name: str
    unqualified_non_ref_type_name: str
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, arg0: str, arg1: str, arg2: bool, arg3: bool) -> None: ...
    def is_left_reference(self) -> bool: ...
    def is_right_reference(self) -> bool: ...

class Typedef:
    name: str
    underlying_name: str
    def __init__(self, *args, **kwargs) -> None: ...

class access_kind:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    private_: ClassVar[access_kind] = ...
    protected_: ClassVar[access_kind] = ...
    public_: ClassVar[access_kind] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class execution_space:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    device: ClassVar[execution_space] = ...
    global_: ClassVar[execution_space] = ...
    host: ClassVar[execution_space] = ...
    host_device: ClassVar[execution_space] = ...
    undefined: ClassVar[execution_space] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class method_kind:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    conversion_function: ClassVar[method_kind] = ...
    copy_constructor: ClassVar[method_kind] = ...
    default_constructor: ClassVar[method_kind] = ...
    destructor: ClassVar[method_kind] = ...
    move_constructor: ClassVar[method_kind] = ...
    other: ClassVar[method_kind] = ...
    other_constructor: ClassVar[method_kind] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class template_param_kind:
    __members__: ClassVar[dict] = ...  # read-only
    __entries: ClassVar[dict] = ...
    non_type: ClassVar[template_param_kind] = ...
    template_: ClassVar[template_param_kind] = ...
    type_: ClassVar[template_param_kind] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

def parse_declarations_from_command_line(
    arg0: list[str], arg1: list[str]
) -> Declarations: ...



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import subprocess
import os
import tempfile
import logging
from typing import Optional

from numba.cuda.cuda_paths import get_nvidia_nvvm_ctk, get_cuda_home

import pylibastcanopy as bindings

from ast_canopy.decl import Function, Struct, ClassTemplate
from ast_canopy.fdcap_min import capture_fd, STREAMFD

logger = logging.getLogger(f"AST_Canopy.{__name__}")


def get_default_cuda_path() -> Optional[str]:
    """Return the path to the default CUDA home directory."""

    # Allow overriding from os.environ
    if home := get_cuda_home():
        return home

    if nvvm_path := get_nvidia_nvvm_ctk():
        # In the form of $ROOT/nvvm/lib64/libnvvm.so, go up 2 levels for cuda home.
        cuda_home = os.path.dirname(os.path.dirname(nvvm_path))

        if os.path.exists(cuda_home):
            logger.info(f"Found CUDA home: {cuda_home}")
            return cuda_home


def get_default_nvcc_path() -> Optional[str]:
    """Return the path to the default NVCC compiler binary."""
    nvvm_path = get_nvidia_nvvm_ctk()

    if not nvvm_path:
        return

    root = os.path.dirname(os.path.dirname(nvvm_path))
    nvcc_path = os.path.join(root, "bin", "nvcc")

    if os.path.exists(nvcc_path):
        logger.info(f"Found NVCC path: {nvcc_path}")
        return nvcc_path


def get_default_compiler_search_paths() -> list[str]:
    """Compile an empty file with clang++ and print logs verbosely.
    Extract the default system header search paths from the logs and return them.
    """

    # Alternatively, use `clang++ --print-search-dirs`
    clang_compile_empty = (
        subprocess.check_output(
            ["clang++", "-E", "-v", "-xc++", "/dev/null"], stderr=subprocess.STDOUT
        )
        .decode()
        .strip()
        .split("\n")
    )
    start = clang_compile_empty.index("#include <...> search starts here:")
    end = clang_compile_empty.index("End of search list.")
    clang_system_header_search_paths = clang_compile_empty[start + 1 : end]
    search_paths = [x.strip() for x in clang_system_header_search_paths]
    logger.info(f"Default compiler search paths: {search_paths=}")
    return search_paths


def get_default_cuda_compiler_include(default="/usr/local/cuda/include") -> str:
    """Compile an empty file with NVCC and extract its default include path.

    ast_canopy depends on a healthy cuda environment to function. If nvcc fails
    to compile an empty CUDA file, this function will raise an error.
    """

    nvcc_bin = get_default_nvcc_path()
    if not nvcc_bin:
        logger.warning("Could not find NVCC binary. Using default nvcc bin from env.")
        nvcc_bin = "nvcc"

    with tempfile.NamedTemporaryFile(suffix=".cu") as tmp_file:
        nvcc_compile_empty = (
            subprocess.run(
                [nvcc_bin, "-E", "-v", tmp_file.name], capture_output=True, check=True
            )
            .stderr.decode()
            .strip()
            .split("\n")
        )

    if s := [i for i in nvcc_compile_empty if "INCLUDES=" in i]:
        include_path = s[0].lstrip("#$ INCLUDES=").strip().strip('"').lstrip("-I")
        logger.info(f"Found NVCC default include path, {include_path=}")
        return include_path
    else:
        logger.warning(
            f"Could not find NVCC default include path. Using default: {default=}"
        )
        return default


def parse_declarations_from_source(
    source_file_path: str,
    files_to_retain: list[str],
    compute_capability: str,
    cccl_root: str = "",
    cudatoolkit_include_dir: str = get_default_cuda_compiler_include(),
    cxx_standard="c++17",
) -> tuple[
    list[Struct],
    list[Function],
    list[bindings.FunctionTemplate],
    list[ClassTemplate],
    list[bindings.Typedef],
    list[bindings.Enum],
]:
    """Given a source file, parse all *top-level* declarations from it.
    Returns a tuple that each contains a list of declaration objects for the source file.

    `files_to_retain` is a required parameter to specify the declarations from which files
    should be ratained in the result. See `Parameters` section for more details.

    Parameters
    ----------
    source_file_path : str
        The path to the source file to parse.

    files_to_retain : list[str]
        A list of file paths, from which the parsed declarations that should be retained in
        the result. A header file usually reference other header files. A semantically intact
        AST should in theory include all referenced header files. In practice, one may not
        need to consume all the other referenced files. To only retain the declarations from
        `source_file_path`, one may pass [source_file_path] to this parameter.

    compute_capability : str
        The compute capability of the target GPU. e.g. "sm_70".

    cccl_root : str, optional
        The root directory of the CCCL project. If not provided, CCCL from default CTK headers
        are used.

    cudatoolkit_include_dir : str, optional
        The path to the CUDA Toolkit include directory. If not provided, the default CUDA include
        directory is used.

    cxx_standard : str, optional
        The C++ standard to use. Default is "c++17".

    Returns
    -------
    tuple:
        A tuple containing lists of declaration objects from the source file.
        See decl.py and pylibastcanopy.pyi for the declaration object types.
    """

    if not os.path.exists(source_file_path):
        raise FileNotFoundError(f"File not found: {source_file_path}")

    if cccl_root:
        cccl_libs = [
            os.path.join(cccl_root, "libcudacxx", "include"),
            os.path.join(cccl_root, "cub"),
            os.path.join(cccl_root, "thrust"),
        ]
        cccl_libs = [f"-I{lib}" for lib in cccl_libs]
    else:
        cccl_libs = []

    clang_resource_file = (
        subprocess.check_output(["clang++", "-print-resource-dir"]).decode().strip()
    )

    clang_search_paths = get_default_compiler_search_paths()

    def custom_cuda_home() -> list[str]:
        cuda_path = get_default_cuda_path()
        if cuda_path:
            return [f"--cuda-path={cuda_path}"]
        else:
            return []

    command_line_options = [
        "clang++",
        "--cuda-device-only",
        "-xcuda",
        f"--cuda-gpu-arch={compute_capability}",
        *custom_cuda_home(),
        f"-std={cxx_standard}",
        f"-isystem{clang_resource_file}/include/",
        *[f"-I{path}" for path in clang_search_paths],
        *cccl_libs,
        f"-I{cudatoolkit_include_dir}",
        source_file_path,
    ]

    logger.debug(f"{command_line_options=}")

    with capture_fd(STREAMFD.STDERR) as cap:
        decls = bindings.parse_declarations_from_command_line(
            command_line_options, files_to_retain
        )

    werr = cap.snap()
    if werr:
        liblogger = logging.getLogger("libastcanopy")
        liblogger.debug(werr)
        if (
            "CUDA version" in werr
            and "is newer than the latest supported version" in werr
        ):
            liblogger.info(
                "Installed cudaToolkit version is newer than the latest supported version of the clangTooling "
                "backend. clangTooling will treat the cudaToolkit as if it is its latest supported version."
            )

    structs = [Struct.from_c_obj(c_obj) for c_obj in decls.records]
    functions = [Function.from_c_obj(c_obj) for c_obj in decls.functions]
    class_templates = [
        ClassTemplate.from_c_obj(c_obj) for c_obj in decls.class_templates
    ]

    return (
        structs,
        functions,
        decls.function_templates,
        class_templates,
        decls.typedefs,
        decls.enums,
    )
